/*
 * UselessStatesRemover.h
 *
 *  Created on: 23. 3. 2014
 *	  Author: Tomas Pecka
 */

#ifndef USELESS_STATES_REMOVER_H_
#define USELESS_STATES_REMOVER_H_

#include <automaton/properties/UsefullStates.h>

#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/CompactNFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/MultiInitialStateNFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/DFA.h>
#include <automaton/TA/DFTA.h>

namespace automaton {

namespace simplify {

class UselessStatesRemover {
public:
	/**
	 * Removes dead states from FSM. Melichar 2.29
	 */
	template<class T, class SymbolType = typename automaton::SymbolTypeOfAutomaton < T >, class StateType = typename automaton::StateTypeOfAutomaton < T > >
	static T remove( const T & automaton );
	template < class SymbolType, class StateType >
	static automaton::DFA < SymbolType, StateType > remove( const automaton::DFA < SymbolType, StateType > & fsm );
	template < class SymbolType, class StateType >
	static automaton::MultiInitialStateNFA < SymbolType, StateType > remove( const automaton::MultiInitialStateNFA < SymbolType, StateType > & fsm );
	template < class SymbolType, class RankType, class StateType >
	static automaton::DFTA < SymbolType, RankType, StateType > remove( const automaton::DFTA < SymbolType, RankType, StateType > & dfta );
};

template < class T, class SymbolType, class StateType >
T UselessStatesRemover::remove( const T & fsm ) {
	// 1.
	ext::set<StateType> Qu = automaton::properties::UsefullStates::usefullStates( fsm );

	// 2.
	T M(fsm.getInitialState());

	for( const auto & a : fsm.getInputAlphabet( ) )
		M.addInputSymbol( a );

	if(Qu.size() == 0) {
		return M;
	}

	for( const auto & q : Qu )
		M.addState( q );

	for( const auto & t : fsm.getTransitions( ) )
		for( const auto & to : t.second )
			if( /* this part is not needed Qu.count(t.first.first) ) && */ Qu.count(to) )
				M.addTransition( t.first.first, t.first.second, to );

	for( const auto & q : fsm.getFinalStates( ) )
		M.addFinalState( q );

	return M;
}

template < class SymbolType, class StateType >
automaton::DFA < SymbolType, StateType > UselessStatesRemover::remove( const automaton::DFA < SymbolType, StateType > & fsm ) {
	// 1.
	ext::set<StateType> Qu = automaton::properties::UsefullStates::usefullStates( fsm );

	// 2.
	automaton::DFA < SymbolType, StateType > M ( fsm.getInitialState () );

	for( const auto & a : fsm.getInputAlphabet( ) )
		M.addInputSymbol( a );

	if(Qu.size() == 0) {
		return M;
	}

	for( const auto & q : Qu )
		M.addState( q );

	for( const auto & t : fsm.getTransitions( ) )
		if( /* this part is not needed Qu.count( t.first.first ) && */ Qu.count( t.second ) )
			M.addTransition( t.first.first, t.first.second, t.second );

	for( const auto & q : fsm.getFinalStates( ) )
		M.addFinalState( q );

	return M;
}

template < class SymbolType, class StateType >
automaton::MultiInitialStateNFA < SymbolType, StateType > UselessStatesRemover::remove( const automaton::MultiInitialStateNFA < SymbolType, StateType > & fsm ) {
	// 1.
	ext::set<StateType> Qu = automaton::properties::UsefullStates::usefullStates( fsm );

	// 2.
	automaton::MultiInitialStateNFA < SymbolType, StateType > M;

	for( const auto & a : fsm.getInputAlphabet( ) )
		M.addInputSymbol( a );

	if(Qu.size() == 0) {
		return M;
	}

	for( const auto & q : Qu )
		M.addState( q );

	for(const auto& init : fsm.getInitialStates()) {
		if( Qu.count(init) )
			M.addInitialState( init );
	}

	for( const auto & t : fsm.getTransitions( ) )
		for( const auto & to : t.second )
			if( Qu.count(to) )
				M.addTransition( t.first.first, t.first.second, to );

	for( const auto & q : fsm.getFinalStates( ) )
		M.addFinalState( q );

	return M;
}

template < class SymbolType, class RankType, class StateType >
automaton::DFTA < SymbolType, RankType, StateType > UselessStatesRemover::remove( const automaton::DFTA < SymbolType, RankType, StateType > & dfta ) {
	automaton::DFTA < SymbolType, RankType, StateType > res;
	res.setInputAlphabet(dfta.getInputAlphabet());
	res.setStates(dfta.getFinalStates());
	res.setFinalStates(dfta.getFinalStates());

	typedef std::pair < const ext::pair < common::ranked_symbol < SymbolType, RankType >, ext::vector < StateType > >, StateType > Transition;

	ext::map<StateType, ext::set<const Transition *>> transitionsToState;
	for(const auto & transition : dfta.getTransitions()) {
		transitionsToState[transition.second].insert(&transition);
	}

	ext::deque<StateType> queue;
	for (const auto & state : dfta.getFinalStates()) {
		queue.push_back(state);
	}

	while(!queue.empty()) {
		ext::set<const Transition *> & transitions = transitionsToState[queue.front()];
		queue.pop_front();
		for (const Transition * transitionPt : transitions) {
			const Transition &transition = *transitionPt;

			for (const auto & state : transition.first.second) {
				if (res.addState(state)) {
					queue.push_back(state);
				}
			}

			res.addTransition(transition.first.first, transition.first.second, transition.second);
		}
	}

	return res;
}

} /* namespace simplify */

} /* namespace automaton */

#endif /* USELESS_STATES_REMOVER_H_ */
