/*
 * AutomataConcatenationEpsilonTransition.h
 *
 *  Created on: 20. 11. 2014
 *	  Author: Tomas Pecka
 */

#ifndef AUTOMATA_CONCATENATION_EPSILON_TRANSITION_H_
#define AUTOMATA_CONCATENATION_EPSILON_TRANSITION_H_

#include <automaton/FSM/EpsilonNFA.h>

namespace automaton {

namespace transform {

/**
 * Concatenates two automata.
 *  - For finite automata A1, A2, we create automaton L accepting L(A1).L(A2)
 */
class AutomataConcatenationEpsilonTransition {
public:
	template < class SymbolType, class EpsilonType = DefaultEpsilonType, class StateType >
	static automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > concatenation(const automaton::DFA < SymbolType, StateType > & first, const automaton::DFA < SymbolType, StateType > & second);
	template < class SymbolType, class EpsilonType = DefaultEpsilonType, class StateType >
	static automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > concatenation(const automaton::NFA < SymbolType, StateType > & first, const automaton::NFA < SymbolType, StateType > & second);
	template < class SymbolType, class EpsilonType, class StateType >
	static automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > concatenation(const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & first, const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & second);
};

template < class SymbolType, class EpsilonType, class StateType >
automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > AutomataConcatenationEpsilonTransition::concatenation(const automaton::DFA < SymbolType, StateType > & first, const automaton::DFA < SymbolType, StateType > & second) {
	static StateType firstDefault ( 1 );
	static StateType secondDefault ( 2 );

	automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > res(StateType(firstDefault, first.getInitialState()));

	for(const auto& symbol : first.getInputAlphabet())
		res.addInputSymbol(symbol);
	for(const auto& symbol : second.getInputAlphabet())
		res.addInputSymbol(symbol);

	for(const auto& q : first.getStates())
		res.addState(StateType(firstDefault, q));
	for(const auto& q : second.getStates())
		res.addState(StateType(secondDefault, q));

	for(const auto& q : second.getFinalStates())
		res.addFinalState(StateType(secondDefault, q));

	for(const auto& t : first.getTransitions())
		res.addTransition(StateType(firstDefault, t.first.first), t.first.second, StateType(firstDefault, t.second));

	for(const auto& t : second.getTransitions())
		res.addTransition(StateType(secondDefault, t.first.first), t.first.second, StateType(secondDefault, t.second));

	for(const auto& q : first.getFinalStates())
		res.addTransition(StateType(firstDefault, q), StateType(secondDefault, second.getInitialState()));

	return res;
}

template < class SymbolType, class EpsilonType, class StateType >
automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > AutomataConcatenationEpsilonTransition::concatenation(const automaton::NFA < SymbolType, StateType > & first, const automaton::NFA < SymbolType, StateType > & second) {
	static StateType firstDefault ( 1 );
	static StateType secondDefault ( 2 );

	automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > res(StateType(firstDefault, first.getInitialState()));

	for(const auto& symbol : first.getInputAlphabet())
		res.addInputSymbol(symbol);
	for(const auto& symbol : second.getInputAlphabet())
		res.addInputSymbol(symbol);

	for(const auto& q : first.getStates())
		res.addState(StateType(firstDefault, q));
	for(const auto& q : second.getStates())
		res.addState(StateType(secondDefault, q));

	for(const auto& q : second.getFinalStates())
		res.addFinalState(StateType(secondDefault, q));

	for(const auto& t : first.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(StateType(firstDefault, t.first.first), t.first.second, StateType(firstDefault, q));

	for(const auto& t : second.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(StateType(secondDefault, t.first.first), t.first.second, StateType(secondDefault, q));

	for(const auto& q : first.getFinalStates())
		res.addTransition(StateType(firstDefault, q), StateType(secondDefault, second.getInitialState()));

	return res;
}

template < class SymbolType, class EpsilonType, class StateType >
automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > AutomataConcatenationEpsilonTransition::concatenation(const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & first, const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & second) {
	static StateType firstDefault ( 1 );
	static StateType secondDefault ( 2 );

	automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > res(StateType(firstDefault, first.getInitialState()));

	for(const auto& symbol : first.getInputAlphabet())
		res.addInputSymbol(symbol);
	for(const auto& symbol : second.getInputAlphabet())
		res.addInputSymbol(symbol);

	for(const auto& q : first.getStates())
		res.addState(StateType(firstDefault, q));
	for(const auto& q : second.getStates())
		res.addState(StateType(secondDefault, q));

	for(const auto& q : second.getFinalStates())
		res.addFinalState(StateType(secondDefault, q));

	for(const auto& t : first.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(StateType(firstDefault, t.first.first), t.first.second, StateType(firstDefault, q));

	for(const auto& t : second.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(StateType(secondDefault, t.first.first), t.first.second, StateType(secondDefault, q));

	for(const auto& q : first.getFinalStates())
		res.addTransition(StateType(firstDefault, q), StateType(secondDefault, second.getInitialState()));

	return res;
}

} /* namespace transform */

} /* namespace automaton */

#endif /* AUTOMATA_CONCATENATION_EPSILON_H_ */
