/*
 * FibonacciHeap.h
 *
 *   Created on: Apr 7, 2016
 *       Author: Jan Broz
 */

#ifndef FIBONACCI_HEAP_INCLUDED
#define FIBONACCI_HEAP_INCLUDED

#include <stdexcept>
#include <cmath>      // maxDegree

/// fibonacci heap used as mergeable priority queue
template< typename elem_t >
class FibonacciHeap {

 public:

	FibonacciHeap( int (* compare)( const elem_t &, const elem_t & ) );
	~FibonacciHeap();

	/// inserts a node with new value into the heap
	void insert( const elem_t & value );
	/// finds the maximum value in the heap
	const elem_t & getMax() const   { return _max->value; }
	/// finds and removes the maximum value from the heap
	elem_t extractMax();
	/// merges this heap with another heap (!! this is a DESTRUCTIVE merge, heap in argument will be cleared !!)
	void mergeWith( FibonacciHeap<elem_t> && that );

	size_t size() const   { return _size; }

 protected:

	struct Node {
		elem_t value;
		unsigned degree;
		bool   mark;
		Node * parent;
		Node * child;
		Node * prev;
		Node * next;
		Node(const elem_t & val, unsigned deg = 0, Node * par = NULL, Node * chld = NULL, Node * pr = NULL, Node * ne = NULL)
		 : value(val), degree(deg), mark(false), parent(par), child(chld), prev(pr), next(ne) {}
	};

	Node *  _max;     ///< pointer to cyclic doubly linked list of trees
	size_t    _size;    ///< count of elements stored in the heap
	int (*  _compare)( const elem_t &, const elem_t & ); ///< user-defined comparator function

 protected:

	/// deletes one linked list of trees
	void deleteTreeList( Node * head );
	/// connects doubly linked lists from 2 fibonacci heaps and returns address of head of the new linked list
	Node * mergeHeaps( Node * max1, Node * max2 );
	/// goes through list of trees and merges trees with same degree
	void consolidate();
	/// links root of tree1 to root of tree2 (tree1 becomes child of tree2)
	Node * linkTreeToTree( Node * root1, Node * root2 );

};


template< typename elem_t >
FibonacciHeap<elem_t>::FibonacciHeap( int (* compare)( const elem_t &, const elem_t & ) )

	: _max( NULL )
	, _size( 0 )
	, _compare( compare )
{
}

template< typename elem_t >
FibonacciHeap<elem_t>::~FibonacciHeap()
{
	deleteTreeList( _max );
}

template< typename elem_t >
void FibonacciHeap<elem_t>::deleteTreeList( Node * head )
{
	if (!head)
		return;

	Node * actual = head;
	Node * next;
	do {
		next = actual->next;
		deleteTreeList( actual->child );
		delete actual;
		actual = next;
	} while (actual != head);
}

template< typename elem_t >
void FibonacciHeap<elem_t>::insert( const elem_t & value )
{
	Node * newNode = new Node( value, 0, NULL, NULL, NULL, NULL );
	newNode->prev = newNode;
	newNode->next = newNode;     // make this node be a single-element cyclic list

	if (!_max) {
		_max = newNode;       // this heap is empty, newNode becomes head of a linked list
		_size = 1;
		return;
	}

	_max = mergeHeaps( _max, newNode );    // link the newNode into the existing linked list

	_size++;
}

template< typename elem_t >
elem_t FibonacciHeap<elem_t>::extractMax()
{
	Node * z = _max;
	if (z != NULL) {
		if (z->child != NULL) {
			Node * z1 = z->child;
			Node * z2 = z1->prev;
			z1->prev = z->prev;
			z1->prev->next = z1;
			z2->next = z;
			z->prev = z2;
			while (z1 != z) {
				z1->parent = NULL;
				z1 = z1->next;
			}
		}
		if (z == z->next)
			_max = NULL;
		else {
			_max = z->next;
			_max->prev = z->prev;
			z->prev->next = _max;
			consolidate();
		}
	}

	_size--;
	elem_t maxVal = z->value;
	delete z;
	return maxVal;
}

template< typename elem_t >
void FibonacciHeap<elem_t>::mergeWith( FibonacciHeap<elem_t> && that )
{
	if (this->_compare != that._compare)    // nodes of these heaps are sorted by different condition
		throw std::logic_error("compare functions aren't equal, unable to merge");

	if (!this->_max) {
		this->_max  = that._max;    // this heap is empty, move the content from the other heap inside
		this->_size = that._size;
		that._max  = NULL;
		that._size = 0;
		return;
	} else if (!that._max) {
		return;                     // the other heap is empty, nothing to do
	}

	this->_max = mergeHeaps( this->_max, that._max );    // link the other heap into this heap
	that._max = NULL;

	this->_size += that._size;
	that._size = 0;
}

template< typename elem_t >
typename FibonacciHeap<elem_t>::Node * FibonacciHeap<elem_t>::mergeHeaps( Node * max1, Node * max2 )
{
	Node * max2prev = max2->prev;

	max2->prev->next = max1;
	max2->prev = max1->prev;    // this magicaly works even if the lists contain only one node
	max1->prev->next = max2;
	max1->prev = max2prev;

	return _compare( max1->value, max2->value ) < 0 ? max2 : max1;
}

template< typename elem_t >
void FibonacciHeap<elem_t>::consolidate()
{
	unsigned maxDegree;
	Node * * rootsByDegree;
	unsigned i, degree;
	Node * actual, * other;

	maxDegree = log2( _size );
	rootsByDegree = new Node * [ maxDegree + 1 ];
	for (i = 0; i <= maxDegree; i++)
		rootsByDegree[i] = NULL;

	actual = _max;
	do {
		degree = actual->degree;
		while (rootsByDegree[ degree ] != NULL) {
			other = rootsByDegree[ degree ];
			if (_compare( actual->value, other->value ) > 0)
				actual = linkTreeToTree( other, actual );
			else
				actual = linkTreeToTree( actual, other );
			rootsByDegree[ degree ] = NULL;
			degree++;
		}
		rootsByDegree[ degree ] = actual;
		actual->mark = true;
		actual = actual->next;
	} while (!actual->mark);

	_max = NULL;
	for (i = 0; i <= maxDegree; i++) {
		if (rootsByDegree[i] != NULL) {
			rootsByDegree[i]->mark = false;
			if (_max == NULL || _compare( rootsByDegree[i]->value, _max->value ) > 0)
				_max = rootsByDegree[i];
		}
	}
	delete [] rootsByDegree;
}

template< typename elem_t >
typename FibonacciHeap<elem_t>::Node * FibonacciHeap<elem_t>::linkTreeToTree( Node * root1, Node * root2 )
{
	root1->prev->next = root1->next;
	root1->next->prev = root1->prev;
	if (!root2->child) {
		root1->next = root1;
		root1->prev = root1;
	} else {
		root1->next = root2->child;
		root1->prev = root2->child->prev;
		root1->next->prev = root1;
		root1->prev->next = root1;
	}
	root2->child = root1;
	root1->parent = root2;
	root2->degree++;
	root1->mark = false;

	return root2;
}


#endif // FIBONACCI_HEAP_INCLUDED
