/*
 * FordFulkerson.cpp
 *
 *   Created on: Mar 29, 2016
 *       Author: Jan Broz
 */

#include "FordFulkerson.h"

#include <unordered_set>  // return value
#include <unordered_map>  // return value
#include <vector>         // removed edges
#include <queue>          // BFS

#include <alib/utility>

#include <exception/CommonException.h>

#include "../maximumflow/FordFulkerson.h"
#include <registration/AlgoRegistration.hpp>

#include <alib/iostream>

namespace graph {

namespace minimumcut {

static graph::maximumflow::Capacity getCapacity( const DirectedGraph & graph )
{
	graph::maximumflow::Capacity capacity;
	int edge_cap;

	for (const DirectedEdge & edge : graph.getEdges())
		capacity[ edge.getFromNode() ][ edge.getToNode() ] = 0;

	for (const DirectedEdge & edge : graph.getEdges()) {
		if (capacity[ edge.getFromNode() ][ edge.getToNode() ] != 0)
			throw exception::CommonException("FordFulkerson: graph is not simple");
		try {
			edge_cap = graph.getEdgeValue( edge );
		} catch (const std::out_of_range & ex) { // when this edge has no value assigned
			throw exception::CommonException("FordFulkerson: an edge is missing a capacity");
		}
		capacity[ edge.getFromNode() ][ edge.getToNode() ] = edge_cap;
	}

	return capacity;
}

static graph::maximumflow::Capacity getCapacity( const UndirectedGraph & graph )
{
	graph::maximumflow::Capacity capacity;
	int edge_cap;

	for (const UndirectedEdge & edge : graph.getEdges()) {
		capacity[ edge.getFirstNode() ][ edge.getSecondNode() ] = 0;
		capacity[ edge.getSecondNode() ][ edge.getFirstNode() ] = 0;
	}

	for (const UndirectedEdge & edge : graph.getEdges()) {
		if (capacity[ edge.getFirstNode() ][ edge.getSecondNode() ] != 0)
			throw exception::CommonException("FordFulkerson: graph is not simple");
		try {
			edge_cap = graph.getEdgeValue( edge );
		} catch (const std::out_of_range & ex) { // when this edge has no value assigned
			throw exception::CommonException("FordFulkerson: an edge is missing a capacity");
		}
		capacity[ edge.getFirstNode() ][ edge.getSecondNode() ] = edge_cap;
		capacity[ edge.getSecondNode() ][ edge.getFirstNode() ] = edge_cap;
	}

	return capacity;
}

static Cut fordfulkerson_impl_dir( const DirectedGraph & graph, Node source, Node sink )
{
	Cut cut;

	graph::maximumflow::Capacity capacity = getCapacity( graph );

	graph::maximumflow::Flow flow = graph::maximumflow::FordFulkerson::fordfulkerson( graph, source, sink );

	// mark nodes, which are reachable from source
	ext::unordered_map<Node,int> state;
	std::queue<Node> open;
	ext::vector<std::pair<Node,Node>> candidates;
	for (const Node & node : graph.getNodes())
		state[ node ] = 0;
	state[ source ] = 1;
	open.push( source );
	while (!open.empty()) {
		Node actual = open.front(); open.pop();
		for (const Node & succ : graph.successors( actual )) {
			if (state[ succ ] == 0 && flow[ actual ][ succ ] < capacity[ actual ][ succ ]) {
				state[ succ ] = 1;
				open.push( succ );
			} else if (flow[ actual ][ succ ] == capacity[ actual ][ succ ]) {
				candidates.push_back({ actual, succ });
			}
		}
		for (const Node & pred : graph.predecessors( actual )) {
			if (state[ pred ] == 0 && flow[ pred ][ actual ] > 0) {
				state[ pred ] = 1;
				open.push( pred );
			} else if (flow[ pred ][ actual ] == 0) {
				candidates.push_back({ pred, actual });
			}
		}
	}

	// cut are those edges, which lead from nodes reachable from source to nodes unreachable from source
	for (std::pair<Node,Node> edge : candidates) {
		if ((state[ edge.first ] == 1 && state[ edge.second ] == 0)
		 || (state[ edge.first ] == 0 && state[ edge.second ] == 1))
			cut.insert( edge );
	}

	return cut;
}

static Cut fordfulkerson_impl_undir( const UndirectedGraph & ugraph, Node source, Node sink )
{
	Cut cut;

	graph::maximumflow::Capacity capacity = getCapacity( ugraph );

	graph::maximumflow::Flow flow = graph::maximumflow::FordFulkerson::fordfulkerson( ugraph, source, sink );

	// mark nodes, which are reachable from source
	ext::unordered_map<Node,int> state;
	std::queue<Node> open;
	ext::vector<std::pair<Node,Node>> candidates;
	for (const Node & node : ugraph.getNodes())
		state[ node ] = 0;
	state[ source ] = 1;
	open.push( source );
	while (!open.empty()) {
		Node actual = open.front(); open.pop();
		for (const Node & neigh : ugraph.neighbors( actual )) {
			if (state[ neigh ] == 0 && flow[ actual ][ neigh ] < capacity[ actual ][ neigh ]) {
				state[ neigh ] = 1;
				open.push( neigh );
			} else if (flow[ actual ][ neigh ] == capacity[ actual ][ neigh ]) {
				candidates.push_back({ actual, neigh });
			}
		}
	}

	// cut are those edges, which lead from nodes reachable from source to nodes unreachable from source
	for (std::pair<Node,Node> edge : candidates) {
		if ((state[ edge.first ] == 1 && state[ edge.second ] == 0)
		 || (state[ edge.first ] == 0 && state[ edge.second ] == 1))
			cut.insert( edge );
	}

	return cut;
}

Cut FordFulkerson::fordfulkerson( const DirectedGraph & graph, Node source, Node sink )
{
	return fordfulkerson_impl_dir( graph, source, sink );
}

Cut FordFulkerson::fordfulkerson( const UndirectedGraph & graph, Node source, Node sink )
{
	return fordfulkerson_impl_undir( graph, source, sink );
}

//auto BasicAlgorithmUndirectedGraph = registration::OverloadRegister < FordFulkerson, graph::UndirectedGraph, graph::UndirectedGraph > ( FordFulkerson::fordfulkerson );

} // namespace minimumcut

} // namespace graph
