/*
 * GrammarToXMLComposer.h
 *
 *  Created on: Nov 11, 2013
 *      Author: Jan Travnicek
 */

#ifndef GRAMMAR_TO_XML_COMPOSER_H_
#define GRAMMAR_TO_XML_COMPOSER_H_

#include <alib/deque>
#include <alib/vector>
#include <alib/tuple>
#include <alib/set>
#include <alib/variant>

#include <sax/Token.h>
#include <core/xmlApi.hpp>

namespace grammar {

/**
 * This class contains methods to print XML representation of automata to the output stream.
 */
class GrammarToXMLComposer {
public:
	template < class SymbolType >
	static void composeNonterminalAlphabet(ext::deque<sax::Token>& out, const ext::set<SymbolType>& symbols);
	template < class SymbolType >
	static void composeTerminalAlphabet(ext::deque<sax::Token>& out, const ext::set<SymbolType>& symbols);
	template < class SymbolType >
	static void composeInitialSymbol(ext::deque<sax::Token>& out, const SymbolType& symbol);

	static void composeGeneratesEpsilon(ext::deque<sax::Token>& out, bool generatesEpsilon) {
		out.emplace_back("generatesEpsilon", sax::Token::TokenType::START_ELEMENT);
		if(generatesEpsilon) {
			out.emplace_back("true", sax::Token::TokenType::START_ELEMENT);
			out.emplace_back("true", sax::Token::TokenType::END_ELEMENT);
		} else {
			out.emplace_back("false", sax::Token::TokenType::START_ELEMENT);
			out.emplace_back("false", sax::Token::TokenType::END_ELEMENT);
		}
		out.emplace_back("generatesEpsilon", sax::Token::TokenType::END_ELEMENT);
	}

	template < class SymbolType >
	static void composeRuleLContext(ext::deque<sax::Token>& out, const ext::vector<SymbolType>& symbols);
	template < class SymbolType >
	static void composeRuleLHS(ext::deque<sax::Token>& out, const ext::vector<SymbolType>& symbols);
	template < class SymbolType >
	static void composeRuleRContext(ext::deque<sax::Token>& out, const ext::vector<SymbolType>& symbols);
	template < class SymbolType >
	static void composeRuleSingleSymbolLHS(ext::deque<sax::Token>& out, const SymbolType& symbol);
	template < class SymbolType >
	static void composeRuleRHS(ext::deque<sax::Token>& out, const ext::vector<SymbolType>& symbols);
	template < class SymbolType >
	static void composeRuleOneOrTwoSymbolsRHS(ext::deque<sax::Token>& out, const ext::variant<SymbolType, ext::pair<SymbolType, SymbolType>>& symbols);
	template < class SymbolType >
	static void composeRuleGNFRHS(ext::deque<sax::Token>& out, const ext::pair<SymbolType, ext::vector<SymbolType>>& symbols);
	template < class SymbolType >
	static void composeRuleLeftLGRHS(ext::deque<sax::Token>& out, const ext::variant<ext::vector<SymbolType>, ext::pair<SymbolType, ext::vector<SymbolType>>>& symbols);
	template < class SymbolType >
	static void composeRuleRightLGRHS(ext::deque<sax::Token>& out, const ext::variant<ext::vector<SymbolType>, ext::pair<ext::vector<SymbolType>, SymbolType>>& symbols);
	template < class SymbolType >
	static void composeRuleLGRHS(ext::deque<sax::Token>& out, const ext::variant<ext::vector<SymbolType>, ext::tuple<ext::vector<SymbolType>, SymbolType, ext::vector<SymbolType>>>& symbols);
};

template < class SymbolType >
void GrammarToXMLComposer::composeNonterminalAlphabet(ext::deque<sax::Token>& out, const ext::set<SymbolType>& symbols) {
	out.emplace_back("nonterminalAlphabet", sax::Token::TokenType::START_ELEMENT);
	for (const auto& symbol : symbols) {
		core::xmlApi<SymbolType>::compose(out, symbol);
	}
	out.emplace_back("nonterminalAlphabet", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void GrammarToXMLComposer::composeTerminalAlphabet(ext::deque<sax::Token>& out, const ext::set<SymbolType>& symbols) {
	out.emplace_back("terminalAlphabet", sax::Token::TokenType::START_ELEMENT);
	for (const auto& symbol : symbols) {
		core::xmlApi<SymbolType>::compose(out, symbol);
	}
	out.emplace_back("terminalAlphabet", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void GrammarToXMLComposer::composeInitialSymbol(ext::deque<sax::Token>& out, const SymbolType& symbol) {
	out.emplace_back("initialSymbol", sax::Token::TokenType::START_ELEMENT);
	core::xmlApi<SymbolType>::compose(out, symbol);
	out.emplace_back("initialSymbol", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void GrammarToXMLComposer::composeRuleLContext(ext::deque<sax::Token>& out, const ext::vector<SymbolType>& symbols) {
	out.emplace_back("lContext", sax::Token::TokenType::START_ELEMENT);
	if(symbols.size() == 0) {
		out.emplace_back("epsilon", sax::Token::TokenType::START_ELEMENT);
		out.emplace_back("epsilon", sax::Token::TokenType::END_ELEMENT);
	} else for (const auto& symbol : symbols) {
		core::xmlApi<SymbolType>::compose(out, symbol);
	}
	out.emplace_back("lContext", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void GrammarToXMLComposer::composeRuleLHS(ext::deque<sax::Token>& out, const ext::vector<SymbolType>& symbols) {
	out.emplace_back("lhs", sax::Token::TokenType::START_ELEMENT);
	if(symbols.size() == 0) {
		out.emplace_back("epsilon", sax::Token::TokenType::START_ELEMENT);
		out.emplace_back("epsilon", sax::Token::TokenType::END_ELEMENT);
	} else for (const auto& symbol : symbols) {
		core::xmlApi<SymbolType>::compose(out, symbol);
	}
	out.emplace_back("lhs", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void GrammarToXMLComposer::composeRuleRContext(ext::deque<sax::Token>& out, const ext::vector<SymbolType>& symbols) {
	out.emplace_back("rContext", sax::Token::TokenType::START_ELEMENT);
	if(symbols.size() == 0) {
		out.emplace_back("epsilon", sax::Token::TokenType::START_ELEMENT);
		out.emplace_back("epsilon", sax::Token::TokenType::END_ELEMENT);
	} else for (const auto& symbol : symbols) {
		core::xmlApi<SymbolType>::compose(out, symbol);
	}
	out.emplace_back("rContext", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void GrammarToXMLComposer::composeRuleSingleSymbolLHS(ext::deque<sax::Token>& out, const SymbolType& symbol) {
	out.emplace_back("lhs", sax::Token::TokenType::START_ELEMENT);
	core::xmlApi<SymbolType>::compose(out, symbol);
	out.emplace_back("lhs", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void GrammarToXMLComposer::composeRuleRHS(ext::deque<sax::Token>& out, const ext::vector<SymbolType>& symbols) {
	out.emplace_back("rhs", sax::Token::TokenType::START_ELEMENT);
	if(symbols.size() == 0) {
		out.emplace_back("epsilon", sax::Token::TokenType::START_ELEMENT);
		out.emplace_back("epsilon", sax::Token::TokenType::END_ELEMENT);
	} else for (const auto& symbol : symbols) {
		core::xmlApi<SymbolType>::compose(out, symbol);
	}
	out.emplace_back("rhs", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void GrammarToXMLComposer::composeRuleOneOrTwoSymbolsRHS(ext::deque<sax::Token>& out, const ext::variant<SymbolType, ext::pair<SymbolType, SymbolType>>& symbols) {
	out.emplace_back("rhs", sax::Token::TokenType::START_ELEMENT);
	if(symbols.template is<SymbolType>()) {
		core::xmlApi<SymbolType>::compose(out, symbols.template get<SymbolType>());
	} else {
		const ext::pair<SymbolType, SymbolType>& rhs = symbols.template get<ext::pair<SymbolType, SymbolType>>();
		core::xmlApi<SymbolType>::compose(out, rhs.first);
		core::xmlApi<SymbolType>::compose(out, rhs.second);
	}
	out.emplace_back("rhs", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void GrammarToXMLComposer::composeRuleGNFRHS(ext::deque<sax::Token>& out, const ext::pair<SymbolType, ext::vector<SymbolType>>& symbols) {
	out.emplace_back("rhs", sax::Token::TokenType::START_ELEMENT);

	core::xmlApi<SymbolType>::compose(out, symbols.first);

	for (const auto& symbol : symbols.second) {
		core::xmlApi<SymbolType>::compose(out, symbol);
	}

	out.emplace_back("rhs", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void GrammarToXMLComposer::composeRuleLGRHS(ext::deque<sax::Token>& out, const ext::variant<ext::vector<SymbolType>, ext::tuple<ext::vector<SymbolType>, SymbolType, ext::vector<SymbolType>>>& symbols) {
	out.emplace_back("rhs", sax::Token::TokenType::START_ELEMENT);
	if(symbols.template is<ext::vector<SymbolType>>()) {
		const ext::vector<SymbolType>& rhs = symbols.template get<ext::vector<SymbolType>>();
		if(rhs.size() == 0) {
			out.emplace_back("epsilon", sax::Token::TokenType::START_ELEMENT);
			out.emplace_back("epsilon", sax::Token::TokenType::END_ELEMENT);
		} else for (const auto& symbol : symbols.template get<ext::vector<SymbolType>>()) {
			core::xmlApi<SymbolType>::compose(out, symbol);
		}
	} else {
		const ext::tuple<ext::vector<SymbolType>, SymbolType, ext::vector<SymbolType>>& rhs = symbols.template get<ext::tuple<ext::vector<SymbolType>, SymbolType, ext::vector<SymbolType>>>();
		for (const auto& symbol : std::get<0>(rhs)) {
			core::xmlApi<SymbolType>::compose(out, symbol);
		}
		core::xmlApi<SymbolType>::compose(out, std::get<1>(rhs));
		for (const auto& symbol : std::get<2>(rhs)) {
			core::xmlApi<SymbolType>::compose(out, symbol);
		}
	}
	out.emplace_back("rhs", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void GrammarToXMLComposer::composeRuleLeftLGRHS(ext::deque<sax::Token>& out, const ext::variant<ext::vector<SymbolType>, ext::pair<SymbolType, ext::vector<SymbolType>>>& symbols) {
	out.emplace_back("rhs", sax::Token::TokenType::START_ELEMENT);
	if(symbols.template is<ext::vector<SymbolType>>()) {
		const ext::vector<SymbolType>& rhs = symbols.template get<ext::vector<SymbolType>>();
		if(rhs.size() == 0) {
			out.emplace_back("epsilon", sax::Token::TokenType::START_ELEMENT);
			out.emplace_back("epsilon", sax::Token::TokenType::END_ELEMENT);
		} else for (const auto& symbol : symbols.template get<ext::vector<SymbolType>>()) {
			core::xmlApi<SymbolType>::compose(out, symbol);
		}
	} else {
		const ext::pair<SymbolType, ext::vector<SymbolType>>& rhs = symbols.template get<ext::pair<SymbolType, ext::vector<SymbolType>>>();
		core::xmlApi<SymbolType>::compose(out, rhs.first);
		for (const auto& symbol : rhs.second) {
			core::xmlApi<SymbolType>::compose(out, symbol);
		}
	}
	out.emplace_back("rhs", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void GrammarToXMLComposer::composeRuleRightLGRHS(ext::deque<sax::Token>& out, const ext::variant<ext::vector<SymbolType>, ext::pair<ext::vector<SymbolType>, SymbolType>>& symbols) {
	out.emplace_back("rhs", sax::Token::TokenType::START_ELEMENT);
	if(symbols.template is<ext::vector<SymbolType>>()) {
		const ext::vector<SymbolType>& rhs = symbols.template get<ext::vector<SymbolType>>();
		if(rhs.size() == 0) {
			out.emplace_back("epsilon", sax::Token::TokenType::START_ELEMENT);
			out.emplace_back("epsilon", sax::Token::TokenType::END_ELEMENT);
		} else {
			for (const auto& symbol : symbols.template get<ext::vector<SymbolType>>()) {
				core::xmlApi<SymbolType>::compose(out, symbol);
			}
		}
	} else {
		const ext::pair<ext::vector<SymbolType>, SymbolType>& rhs = symbols.template get<ext::pair<ext::vector<SymbolType>, SymbolType>>();
		for (const auto& symbol : rhs.first) {
			core::xmlApi<SymbolType>::compose(out, symbol);
		}
		core::xmlApi<SymbolType>::compose(out, rhs.second);
	}
	out.emplace_back("rhs", sax::Token::TokenType::END_ELEMENT);
}

} /* namespace grammar */

#endif /* GRAMMAR_TO_XML_COMPOSER_H_ */

