/*
 * SuffixArray.h
 *
 *  Created on: Jan 8, 2017
 *      Author: Jan Travnicek
 */

#ifndef SUFFIX_ARRAY_H_
#define SUFFIX_ARRAY_H_

#include <alib/string>
#include <alib/set>
#include <alib/trie>
#include <alib/variant>
#include <alib/iostream>
#include <alib/algorithm>
#include <sstream>

#include <common/DefaultSymbolType.h>

#include <core/components.hpp>
#include <exception/CommonException.h>

#include <object/UniqueObject.h>
#include <object/ObjectBase.h>

#include <sax/FromXMLParserHelper.h>
#include <core/xmlApi.hpp>
#include <primitive/Unsigned.h>
#include <primitive/xml/Unsigned.h>

#include <container/ObjectsSet.h>
#include <container/ObjectsVector.h>

#include <container/xml/ObjectsSet.h>
#include <container/xml/ObjectsVector.h>

#include <alphabet/common/SymbolNormalize.h>

namespace indexes {

namespace stringology {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
template < class SymbolType = DefaultSymbolType >
class SuffixArray final : public object::ObjectBase, public core::Components < SuffixArray < SymbolType >, ext::set < SymbolType >, component::Set, GeneralAlphabet > {
protected:
	ext::vector < unsigned > m_data;
	ext::vector < SymbolType > m_string;

public:
	/**
	 * @copydoc SuffixTrieNode::clone() const
	 */
	virtual ObjectBase * clone ( ) const;

	/**
	 * @copydoc SuffixTrieNode::plunder() const
	 */
	virtual ObjectBase * plunder ( ) &&;

	explicit SuffixArray ( ext::set < SymbolType > alphabet, ext::vector < unsigned > data, ext::vector < SymbolType > string );

	/**
	 * @return Root node of the trie
	 */
	const ext::vector < unsigned > & getData ( ) const &;

	ext::vector < unsigned > && getData ( ) &&;

	const ext::vector < SymbolType > & getString ( ) const &;

	ext::vector < SymbolType > && getString ( ) &&;

	const ext::set < SymbolType > & getAlphabet ( ) const & {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	ext::set < SymbolType > && getAlphabet ( ) && {
		return std::move ( this->template accessComponent < GeneralAlphabet > ( ).get ( ) );
	}

	/**
	 * Sets the root node of the regular expression tree
	 * @param tree root node to set
	 */
	void setData ( ext::vector < unsigned > data );

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromAlphabet ( const SymbolType & symbol ) {
		return this->template accessComponent < GeneralAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const SuffixArray & other ) const;

	virtual explicit operator std::string ( ) const;

	virtual object::ObjectBase * inc ( ) &&;

	typedef SuffixArray < > normalized_type;
};

} /* namespace stringology */

} /* namespace indexes */

namespace indexes {

namespace stringology {

template < class SymbolType >
SuffixArray < SymbolType >::SuffixArray ( ext::set < SymbolType > alphabet, ext::vector < unsigned > data, ext::vector < SymbolType > string ) : core::Components < SuffixArray, ext::set < SymbolType >, component::Set, GeneralAlphabet > ( std::move ( alphabet ) ), m_data ( std::move ( data ) ), m_string ( std::move ( string ) ) {
	// TODO check validity of the string like in LinearString
}

template < class SymbolType >
object::ObjectBase * SuffixArray < SymbolType >::clone ( ) const {
	return new SuffixArray ( * this );
}

template < class SymbolType >
object::ObjectBase * SuffixArray < SymbolType >::plunder ( ) && {
	return new SuffixArray ( std::move ( * this ) );
}

template < class SymbolType >
const ext::vector < unsigned > & SuffixArray < SymbolType >::getData ( ) const & {
	return m_data;
}

template < class SymbolType >
ext::vector < unsigned > && SuffixArray < SymbolType >::getData ( ) && {
	return std::move ( m_data );
}

template < class SymbolType >
const ext::vector < SymbolType > & SuffixArray < SymbolType >::getString ( ) const & {
	return m_string;
}

template < class SymbolType >
ext::vector < SymbolType > && SuffixArray < SymbolType >::getString ( ) && {
	return std::move ( m_string );
}

template < class SymbolType >
void SuffixArray < SymbolType >::setData ( ext::vector < unsigned > data ) {
	this->m_data = std::move ( data );
}

template < class SymbolType >
void SuffixArray < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(SuffixArray " << this->m_data << ", " << this->m_string << ")";
}

template < class SymbolType >
int SuffixArray < SymbolType >::compare ( const SuffixArray & other ) const {
	auto first = ext::tie ( getData ( ), getString ( ), getAlphabet ( ) );
	auto second = ext::tie ( other.getData ( ), other.getString ( ), other.getAlphabet ( ) );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
SuffixArray < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType >
object::ObjectBase* SuffixArray < SymbolType >::inc() && {
	return new object::UniqueObject(object::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace stringology */

} /* namespace indexes */

namespace core {

template < class SymbolType >
class SetConstraint < indexes::stringology::SuffixArray < SymbolType >, SymbolType, indexes::stringology::GeneralAlphabet > {
public:
	static bool used ( const indexes::stringology::SuffixArray < SymbolType > & index, const SymbolType & symbol ) {
		const ext::vector < SymbolType > & content = index.getString ( );
		return std::find ( content.begin(), content.end(), symbol ) != content.end();
	}

	static bool available ( const indexes::stringology::SuffixArray < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const indexes::stringology::SuffixArray < SymbolType > &, const SymbolType & ) {
	}
};

template < class SymbolType >
struct normalize < indexes::stringology::SuffixArray < SymbolType >, typename std::enable_if < ! std::is_same < indexes::stringology::SuffixArray < SymbolType >, indexes::stringology::SuffixArray < > >::value >::type > {
	static indexes::stringology::SuffixArray < > eval ( indexes::stringology::SuffixArray < SymbolType > && value ) {
		ext::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( value ).getAlphabet ( ) );
		ext::vector < DefaultSymbolType > string = alphabet::SymbolNormalize::normalizeSymbols ( std::move ( value ).getString ( ) );

		return indexes::stringology::SuffixArray < > ( std::move ( alphabet ), std::move ( value ).getData ( ), std::move ( string ) );
	}
};

template < class SymbolType >
struct xmlApi < indexes::stringology::SuffixArray < SymbolType > > {
	static indexes::stringology::SuffixArray < SymbolType > parse ( ext::deque < sax::Token >::iterator & input );
	static bool first ( const ext::deque < sax::Token >::const_iterator & input );
	static const std::string & xmlTagName ( );
	static void compose ( ext::deque < sax::Token > & output, const indexes::stringology::SuffixArray < SymbolType > & data );
};

template < class SymbolType >
indexes::stringology::SuffixArray < SymbolType > xmlApi < indexes::stringology::SuffixArray < SymbolType > >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
	ext::set < SymbolType > alphabet = core::xmlApi < ext::set < SymbolType > >::parse ( input );
	ext::vector < unsigned > data = core::xmlApi < ext::vector < unsigned > >::parse ( input );
	ext::vector < SymbolType > string = core::xmlApi < ext::vector < SymbolType > >::parse ( input );
	indexes::stringology::SuffixArray < SymbolType > res ( std::move ( alphabet ), std::move ( data ), std::move ( string ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, xmlTagName ( ) );
	return res;
}

template < class SymbolType >
bool xmlApi < indexes::stringology::SuffixArray < SymbolType > >::first ( const ext::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
}

template < class SymbolType >
const std::string & xmlApi < indexes::stringology::SuffixArray < SymbolType > >::xmlTagName ( ) {
	static std::string xmlTagName = "SuffixArray";

	return xmlTagName;
}

template < class SymbolType >
void xmlApi < indexes::stringology::SuffixArray < SymbolType > >::compose ( ext::deque < sax::Token > & output, const indexes::stringology::SuffixArray < SymbolType > & index ) {
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::START_ELEMENT );
	core::xmlApi < ext::set < SymbolType > >::compose ( output, index.getAlphabet ( ) );
	core::xmlApi < ext::vector < unsigned > >::compose ( output, index.getData ( ) );
	core::xmlApi < ext::vector < SymbolType > >::compose ( output, index.getString ( ) );
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace core */

#endif /* SUFFIX_ARRAY_H_ */
