/*
 * FormalRTE.h
 *
 *  Created on: Feb 20, 2016
 *      Author: Tomas Pecka
 */

#ifndef FORMAL_RTE_H_
#define FORMAL_RTE_H_

#include <alib/string>
#include <alib/set>
#include <alib/iostream>
#include <alib/algorithm>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>
#include <exception/CommonException.h>

#include "../RTEBase.h"
#include "../RTEFeatures.h"
#include "../common/RTEFromXMLParser.h"
#include "../common/RTEToXMLComposer.h"

#include "FormalRTEStructure.h"

#include <core/normalize.hpp>

namespace rte {

class GeneralAlphabet;
class ConstantAlphabet;

/**
 * Represents regular tree expression parsed from the XML. Regular tree expression is stored
 * as a tree of RTEElement elements.
 */
template < class SymbolType, class RankType >
class FormalRTE final : public RTEBase, public core::Components < FormalRTE < SymbolType, RankType >, ext::set < common::ranked_symbol < SymbolType, RankType > >, component::Set, std::tuple < GeneralAlphabet, ConstantAlphabet > > {
protected:
	FormalRTEStructure < SymbolType, RankType > m_rte;

public:
	explicit FormalRTE ( );

	explicit FormalRTE ( ext::set < common::ranked_symbol < SymbolType, RankType > > alphabetF, ext::set < common::ranked_symbol < SymbolType, RankType > > alphabetK, FormalRTEStructure < SymbolType, RankType > rte );
	explicit FormalRTE ( std::pair < ext::set < common::ranked_symbol < SymbolType, RankType > >, ext::set < common::ranked_symbol < SymbolType, RankType > > > alphabets, FormalRTEStructure < SymbolType, RankType > rte );
	explicit FormalRTE ( FormalRTEStructure < SymbolType, RankType > rte );

	/**
	 * @copydoc FormalRTEStructure::clone() const
	 */
	virtual RTEBase * clone ( ) const;

	/**
	 * @copydoc FormalRTEStructure::plunder() const
	 */
	virtual RTEBase * plunder ( ) &&;

	const ext::set < common::ranked_symbol < SymbolType, RankType > > & getAlphabet ( ) const & {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	ext::set < common::ranked_symbol < SymbolType, RankType > > && getAlphabet ( ) && {
		return std::move ( this->template accessComponent < GeneralAlphabet > ( ).get ( ) );
	}

	void addAlphabetSymbols ( common::ranked_symbol < SymbolType, RankType > symbol ) {
		this->template accessComponent < GeneralAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addAlphabetSymbols ( ext::set < common::ranked_symbol < SymbolType, RankType > > symbols ) {
		this->template accessComponent < GeneralAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setAlphabetSymbols ( ext::set < common::ranked_symbol < SymbolType, RankType > > symbols ) {
		this->template accessComponent < GeneralAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeAlphabetSymbol ( const common::ranked_symbol < SymbolType, RankType > & symbol ) {
		this->template accessComponent < GeneralAlphabet > ( ).remove ( symbol );
	}

	const ext::set < common::ranked_symbol < SymbolType, RankType > > & getSubstitutionAlphabet ( ) const & {
		return this->template accessComponent < ConstantAlphabet > ( ).get ( );
	}

	ext::set < common::ranked_symbol < SymbolType, RankType > > && getSubstitutionAlphabet ( ) && {
		return std::move ( this->template accessComponent < ConstantAlphabet > ( ).get ( ) );
	}

	void addConstantSymbol ( common::ranked_symbol < SymbolType, RankType > symbol ) {
		this->template accessComponent < ConstantAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addConstantSymbols ( ext::set < common::ranked_symbol < SymbolType, RankType > > symbols ) {
		this->template accessComponent < ConstantAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setConstantSymbols ( ext::set < common::ranked_symbol < SymbolType, RankType > > symbols ) {
		this->template accessComponent < ConstantAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeConstantSymbol ( const common::ranked_symbol < SymbolType, RankType > & symbol ) {
		this->template accessComponent < ConstantAlphabet > ( ).remove ( symbol );
	}

	/**
	 * @return Root node of the regular expression tree
	 */
	const FormalRTEStructure < SymbolType, RankType > & getRTE ( ) const &;

	/**
	 * @return Root node of the regular expression tree
	 */
	FormalRTEStructure < SymbolType, RankType > && getRTE ( ) &&;

	/**
	 * Sets the root node of the regular expression tree
	 * @param regExp root node to set
	 */
	void setRTE ( FormalRTEStructure < SymbolType, RankType > regExp );

	/**
	 * Prints XML representation of the RTE to the output stream.
	 * @param out output stream to which print the RTE
	 * @param regexp RTE to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const FormalRTE & other ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName ( ) {
		static std::string xmlTagName = "FormalRTE";

		return xmlTagName;
	}

	static FormalRTE parse ( ext::deque < sax::Token >::iterator & input );

	static void compose ( ext::deque < sax::Token > & out, const FormalRTE & regexp );

	virtual object::ObjectBase * inc ( ) &&;

	typedef FormalRTE < > normalized_type;
};

template < class SymbolType, class RankType >
FormalRTE < SymbolType, RankType >::FormalRTE ( ext::set < common::ranked_symbol < SymbolType, RankType > > alphabetF, ext::set < common::ranked_symbol < SymbolType, RankType > > alphabetK, FormalRTEStructure < SymbolType, RankType > rte ) : core::Components < FormalRTE, ext::set < common::ranked_symbol < SymbolType, RankType > >, component::Set, std::tuple < GeneralAlphabet, ConstantAlphabet > > ( std::move ( alphabetF ), std::move ( alphabetK ) ), m_rte ( std::move ( rte ) ) {
	if ( !this->m_rte.getStructure ( ).checkAlphabet ( getAlphabet ( ), getSubstitutionAlphabet ( ) ) )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

template < class SymbolType, class RankType >
FormalRTE < SymbolType, RankType >::FormalRTE ( std::pair < ext::set < common::ranked_symbol < SymbolType, RankType > >, ext::set < common::ranked_symbol < SymbolType, RankType > > > alphabets, FormalRTEStructure < SymbolType, RankType > rte ) : FormalRTE ( std::move ( alphabets.first ), std::move ( alphabets.second ), std::move ( rte ) ) {
}

template < class SymbolType, class RankType >
FormalRTE < SymbolType, RankType >::FormalRTE ( ) : FormalRTE ( ext::set < common::ranked_symbol < SymbolType, RankType > > ( ), ext::set < common::ranked_symbol < SymbolType, RankType > > ( ), FormalRTEStructure < SymbolType, RankType > ( ) ) {
}

template < class SymbolType, class RankType >
FormalRTE < SymbolType, RankType >::FormalRTE ( FormalRTEStructure < SymbolType, RankType > rte ) : FormalRTE ( rte.getStructure ( ).computeMinimalAlphabets ( ), rte ) {
}

template < class SymbolType, class RankType >
RTEBase * FormalRTE < SymbolType, RankType >::clone ( ) const {
	return new FormalRTE ( * this );
}

template < class SymbolType, class RankType >
RTEBase * FormalRTE < SymbolType, RankType >::plunder ( ) && {
	return new FormalRTE ( std::move ( * this ) );
}

template < class SymbolType, class RankType >
const FormalRTEStructure < SymbolType, RankType > & FormalRTE < SymbolType, RankType >::getRTE ( ) const & {
	return m_rte;
}

template < class SymbolType, class RankType >
FormalRTEStructure < SymbolType, RankType > && FormalRTE < SymbolType, RankType >::getRTE ( ) && {
	return m_rte;
}

template < class SymbolType, class RankType >
void FormalRTE < SymbolType, RankType >::setRTE ( FormalRTEStructure < SymbolType, RankType > param ) {
	if ( !param.getStructure ( ).checkAlphabet ( getAlphabet ( ), getSubstitutionAlphabet ( ) ) )
		throw exception::CommonException ( "Symbols not matching alphabets." );

	this->m_rte = std::move ( param );
}

template < class SymbolType, class RankType >
void FormalRTE < SymbolType, RankType >::operator >>( std::ostream & out ) const {
	out << "(FormalRTE " << this->m_rte << ")";
}

template < class SymbolType, class RankType >
int FormalRTE < SymbolType, RankType >::compare ( const FormalRTE & other ) const {
	auto first = ext::tie ( m_rte, getAlphabet ( ), getSubstitutionAlphabet ( ) );
	auto second = ext::tie ( m_rte, other.getAlphabet ( ), getSubstitutionAlphabet ( ) );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType, class RankType >
FormalRTE < SymbolType, RankType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType, class RankType >
FormalRTE < SymbolType, RankType > FormalRTE < SymbolType, RankType >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, FormalRTE::getXmlTagName ( ) );

	std::pair < ext::set < common::ranked_symbol < SymbolType, RankType > >, ext::set < common::ranked_symbol < SymbolType, RankType > > > alphabets = RTEFromXMLParser::parseAlphabet < SymbolType, RankType > ( input );
	FormalRTEStructure < SymbolType, RankType > element ( core::xmlApi < rte::FormalRTEStructure < SymbolType, RankType > >::parse ( input ) );
	FormalRTE < SymbolType, RankType > rte ( std::move ( alphabets.first ), std::move ( alphabets.second ), std::move ( element ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, FormalRTE::getXmlTagName ( ) );
	return rte;
}

template < class SymbolType, class RankType >
void FormalRTE < SymbolType, RankType >::compose ( ext::deque < sax::Token > & out, const FormalRTE & regexp ) {
	out.emplace_back ( FormalRTE::getXmlTagName ( ), sax::Token::TokenType::START_ELEMENT );
	RTEToXMLComposer::composeAlphabet ( out, regexp.getAlphabet ( ), regexp.getSubstitutionAlphabet ( ) );
	core::xmlApi < FormalRTEStructure < SymbolType, RankType > >::compose ( out, regexp.getRTE ( ) );
	out.emplace_back ( FormalRTE::getXmlTagName ( ), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType, class RankType >
object::ObjectBase* FormalRTE < SymbolType, RankType >::inc() && {
	return new object::UniqueObject(object::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace rte */

namespace core {

template < class SymbolType, class RankType >
class SetConstraint< rte::FormalRTE < SymbolType, RankType >, common::ranked_symbol < SymbolType, RankType >, rte::GeneralAlphabet > {
public:
	static bool used ( const rte::FormalRTE < SymbolType, RankType > & rte, const common::ranked_symbol < SymbolType, RankType > & symbol ) {
		return rte.getRTE ( ).getStructure ( ).testSymbol ( symbol );
	}

	static bool available ( const rte::FormalRTE < SymbolType, RankType > &, const common::ranked_symbol < SymbolType, RankType > & ) {
		return true;
	}

	static void valid ( const rte::FormalRTE < SymbolType, RankType > & rte, const common::ranked_symbol < SymbolType, RankType > & symbol ) {
		if ( rte.template accessComponent < rte::ConstantAlphabet > ( ).get ( ).count ( symbol ) )
			throw ::exception::CommonException ( "Symbol " + ext::to_string ( symbol ) + " cannot be in general alphabet since it is already in constant alphabet" );
	}
};

template < class SymbolType, class RankType >
class SetConstraint< rte::FormalRTE < SymbolType, RankType >, common::ranked_symbol < SymbolType, RankType >, rte::ConstantAlphabet > {
public:
	static bool used ( const rte::FormalRTE < SymbolType, RankType > & rte, const common::ranked_symbol < SymbolType, RankType > & symbol ) {
		return rte.getRTE ( ).getStructure ( ).testSymbol ( symbol );
	}

	static bool available ( const rte::FormalRTE < SymbolType, RankType > &, const common::ranked_symbol < SymbolType, RankType > & ) {
		return true;
	}

	static void valid ( const rte::FormalRTE < SymbolType, RankType > & rte, const common::ranked_symbol < SymbolType, RankType > & symbol ) {
		if ( rte.template accessComponent < rte::GeneralAlphabet > ( ).get ( ).count ( symbol ) )
			throw ::exception::CommonException ( "Symbol " + ext::to_string ( symbol ) + " cannot be in constant alphabet since it is already in general alphabet" );
	}
};

template < class SymbolType, class RankType >
struct normalize < rte::FormalRTE < SymbolType, RankType >, typename std::enable_if < ! std::is_same < rte::FormalRTE < SymbolType, RankType >, rte::FormalRTE < > >::value >::type > {
	static rte::FormalRTE < > eval ( rte::FormalRTE < SymbolType, RankType > && value ) {
		ext::set < common::ranked_symbol < DefaultSymbolType, DefaultRankType > > alphabet = alphabet::SymbolNormalize::normalizeRankedAlphabet ( std::move ( value ).getAlphabet ( ) );
		ext::set < common::ranked_symbol < DefaultSymbolType, DefaultRankType > > constants = alphabet::SymbolNormalize::normalizeRankedAlphabet ( std::move ( value ).getSubstitutionAlphabet ( ) );

		return rte::FormalRTE < > ( alphabet, constants, std::move ( value ).getRTE ( ).normalize ( ) );
	}
};

} /* namespace core */

#endif /* FORMAL_RTE_H_ */
