/*
 * CyclicString.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef CYCLIC_STRING_H_
#define CYCLIC_STRING_H_

#include <alib/set>
#include <alib/vector>
#include <sstream>
#include <alib/algorithm>

#include <core/components.hpp>
#include <object/UniqueObject.h>

#include "StringBase.h"
#include "StringFeatures.h"
#include <exception/CommonException.h>
#include <alphabet/common/SymbolNormalize.h>

#include <core/normalize.hpp>

namespace string {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of CyclicStringElement.
 */
template < class SymbolType >
class CyclicString final : public StringBase, public core::Components < CyclicString < SymbolType >, ext::set < SymbolType >, component::Set, GeneralAlphabet > {
	ext::vector < SymbolType > m_Data;

public:
	explicit CyclicString ( );
	explicit CyclicString ( ext::set < SymbolType > alphabet, ext::vector < SymbolType > str );
	explicit CyclicString ( ext::vector < SymbolType > str );
	explicit CyclicString ( const std::string & str );
	explicit CyclicString ( const char * str );
	explicit CyclicString ( const Epsilon < SymbolType > & epsilon );

	virtual StringBase * clone ( ) const;
	virtual StringBase * plunder ( ) &&;

	const ext::set < SymbolType > & getAlphabet ( ) const & {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	ext::set < SymbolType > && getAlphabet ( ) && {
		return std::move ( this->template accessComponent < GeneralAlphabet > ( ).get ( ) );
	}

	void extendAlphabet ( ext::set < SymbolType > symbols ) {
		this->template accessComponent < GeneralAlphabet > ( ).add ( std::move ( symbols ) );
	}

	/**
	 * @return List of symbols forming string (const version).
	 */
	const ext::vector < SymbolType > & getContent ( ) const &;

	ext::vector < SymbolType > && getContent ( ) &&;

	void setContent ( ext::vector < SymbolType > str );

	/**
	 * @return true if string is an empty word (vector length is 0)
	 */
	bool isEmpty ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const CyclicString & other ) const;

	virtual void operator >>( std::ostream & out ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "CyclicString";

		return xmlTagName;
	}

	virtual object::ObjectBase * inc ( ) &&;

	typedef CyclicString < > normalized_type;
};

} /* namespace string */

#include "Epsilon.h"

namespace string {

template < class SymbolType >
CyclicString < SymbolType >::CyclicString(ext::set<SymbolType> alphabet, ext::vector<SymbolType> str) : core::Components < CyclicString, ext::set < SymbolType >, component::Set, GeneralAlphabet > ( std::move ( alphabet ) ) {
	setContent(std::move(str));
}

template < class SymbolType >
CyclicString < SymbolType >::CyclicString() : CyclicString ( ext::set < SymbolType > ( ), ext::vector < SymbolType > ( ) ) {
}

template < class SymbolType >
CyclicString < SymbolType >::CyclicString(ext::vector<SymbolType> str) : CyclicString ( ext::set < SymbolType> ( str.begin(), str.end() ), str ) {
}

template < class SymbolType >
CyclicString < SymbolType >::CyclicString(const std::string& str) : CyclicString ( ext::vector < SymbolType > ( str.begin ( ), str.end ( ) ) ) {
}

template < class SymbolType >
CyclicString < SymbolType >::CyclicString(const char* str) : CyclicString((std::string) str) {
}

template < class SymbolType >
CyclicString < SymbolType >::CyclicString(const Epsilon < SymbolType > & epsilon) : CyclicString ( epsilon.getAlphabet( ), ext::vector < SymbolType > ( ) ) {
}

template < class SymbolType >
StringBase* CyclicString < SymbolType >::clone() const {
	return new CyclicString(*this);
}

template < class SymbolType >
StringBase* CyclicString < SymbolType >::plunder() && {
	return new CyclicString(std::move(*this));
}

template < class SymbolType >
const ext::vector < SymbolType > & CyclicString < SymbolType >::getContent ( ) const & {
	return this->m_Data;
}

template < class SymbolType >
ext::vector < SymbolType > && CyclicString < SymbolType >::getContent ( ) && {
	return std::move ( this->m_Data );
}

//serves as both move and copy content setter
template < class SymbolType >
void CyclicString < SymbolType >::setContent(ext::vector<SymbolType> str) {
	ext::set<SymbolType> minimalAlphabet(str.begin(), str.end());
	ext::set<SymbolType> unknownSymbols;
	std::set_difference(minimalAlphabet.begin(), minimalAlphabet.end(), getAlphabet().begin(), getAlphabet().end(), std::inserter(unknownSymbols, unknownSymbols.end()));

	if(unknownSymbols.size() > 0)
		throw exception::CommonException("Input symbols not in the alphabet.");

	m_Data = std::move(str);
}

template < class SymbolType >
bool CyclicString < SymbolType >::isEmpty() const {
	return this->m_Data.size() == 0;
}

template < class SymbolType >
int CyclicString < SymbolType >::compare(const CyclicString& other) const {
	auto first = ext::tie ( m_Data, getAlphabet ( ) );
	auto second = ext::tie ( other.m_Data, other.getAlphabet ( ) );

	static ext::compare<decltype(first)> comp;
	return comp(first, second);
}

template < class SymbolType >
void CyclicString < SymbolType >::operator >>(std::ostream& out) const {
	out << "(CyclicString ";
	for(const SymbolType& symbol : this->m_Data)
		out << symbol;
	out << ")";
}

template < class SymbolType >
CyclicString < SymbolType >::operator std::string () const {
	std::stringstream ss;
	ss << "<";
	for(const SymbolType& symbol : this->m_Data)
		ss << symbol;
	ss << ">";
	return std::move(ss).str();
}

template < class SymbolType >
object::ObjectBase* CyclicString < SymbolType >::inc() && {
	return new object::UniqueObject(object::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace string */

namespace core {

template < class SymbolType >
class SetConstraint< string::CyclicString < SymbolType >, SymbolType, string::GeneralAlphabet > {
public:
	static bool used ( const string::CyclicString < SymbolType > & str, const SymbolType & symbol ) {
		const ext::vector<SymbolType>& content = str.getContent ( );
		return std::find(content.begin(), content.end(), symbol) != content.end();
	}

	static bool available ( const string::CyclicString < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const string::CyclicString < SymbolType > &, const SymbolType & ) {
	}
};

template < class SymbolType >
struct normalize < string::CyclicString < SymbolType >, typename std::enable_if < ! std::is_same < string::CyclicString < SymbolType >, string::CyclicString < > >::value >::type > {
	static string::CyclicString < > eval ( string::CyclicString < SymbolType > && value ) {
		ext::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( value ).getAlphabet ( ) );
		ext::vector < DefaultSymbolType > content = alphabet::SymbolNormalize::normalizeSymbols ( std::move ( value ).getContent ( ) );
		return string::CyclicString < > ( std::move ( alphabet ), std::move ( content ) );
	}
};

} /* namespace core */

#endif /* CYCLIC_STRING_H_ */
