#include "TreeTest.h"

#include "sax/SaxParseInterface.h"
#include "sax/SaxComposeInterface.h"

#include "tree/ranked/RankedTree.h"
#include "tree/unranked/UnrankedTree.h"
#include "tree/ranked/PrefixRankedTree.h"
#include "tree/ranked/PrefixRankedBarTree.h"
#include "tree/ranked/RankedPattern.h"

#include "tree/TreeException.h"

#include "factory/XmlDataFactory.hpp"

#include "common/ranked_symbol.hpp"
#include "alphabet/SubtreeWildcardSymbol.h"
#include "alphabet/BarSymbol.h"

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION( TreeTest, "tree" );
CPPUNIT_TEST_SUITE_REGISTRATION( TreeTest );

void TreeTest::setUp() {
}

void TreeTest::tearDown() {
}

ext::tree < common::ranked_symbol < > > prefixToNode(const std::string & s, int & pos) {
	char c = s[pos];
	int r = s[pos+1] - '0';
	pos += 2;
	ext::vector<ext::tree < common::ranked_symbol < > >> children;
	for (int i = 0; i < r; i++) {
		children.push_back(prefixToNode(s, pos));
	}
	return ext::tree < common::ranked_symbol < > >(common::ranked_symbol < >(c, r), std::move(children));
}

void TreeTest::testRankedTreeParser() {
/*
	L=(a(2), b(1), c(0))

	\-1.a(2)
	  |
	  |-2.b(1)
	  | |
	  | \-3.c(0)
	  |
	  \-4.c(0)
*/
	const common::ranked_symbol < > a ('a', 2);
	const common::ranked_symbol < > b ('b', 1);
	const common::ranked_symbol < > c ('c', 0);

	const ext::set<common::ranked_symbol < >> alphabet {a, b, c};

	ext::tree < common::ranked_symbol < > > node3(c, {});
	ext::tree < common::ranked_symbol < > > node4(c, {});
	ext::tree < common::ranked_symbol < > > node2(b, {std::move(node3)});
	ext::tree < common::ranked_symbol < > > node1(a, {std::move(node2), std::move(node4)});

	tree::RankedTree < > tree(alphabet, std::move(node1));

	CPPUNIT_ASSERT( tree == tree );
	tree.getContent().nicePrint(std::cout);
	{
		ext::deque<sax::Token> tokens = factory::XmlDataFactory::toTokens(tree);
		std::string tmp;
		sax::SaxComposeInterface::printMemory(tmp, tokens);

		ext::deque<sax::Token> tokens2;
		sax::SaxParseInterface::parseMemory(tmp, tokens2);
		tree::RankedTree < > tree2 = factory::XmlDataFactory::fromTokens (std::move(tokens2));

		CPPUNIT_ASSERT( tree == tree2 );
		std::cout << std::endl;
		tree2.getContent().nicePrint(std::cout);
	}

	std::string s = "a2a2a2b1a2c0c0c0c0c0";
	int itmp = 0;
	ext::tree < common::ranked_symbol < > > node = prefixToNode(s, itmp);
	ext::set<common::ranked_symbol < >> al;
	for (unsigned i = 0; i < s.length(); i += 2) al.insert(common::ranked_symbol < >(s[i], (int) (s[i+1] - '0')));
	tree::RankedTree < > t(al, std::move(node));
	factory::XmlDataFactory::toStdout(t);

	//TODO test it somehow
}

void TreeTest::testRankedTreeCompare() {
/*
	L=(a(2), b(1), c(0))

	\-1.a(2)
	  |
	  |-2.b(1)
	  | |
	  | \-3.c(0)
	  |
	  \-4.c(0)

	\-5.a(2)
	  |
	  |-6.c(0)
	  |
	  \-7.b(1)
	    |
	    \-8.c(0)
*/
	const common::ranked_symbol < > a ('a', 2);
	const common::ranked_symbol < > b ('b', 1);
	const common::ranked_symbol < > c ('c', 0);

	const ext::set<common::ranked_symbol < >> alphabet {a, b, c};

	ext::tree < common::ranked_symbol < > > node3(c, {});
	ext::tree < common::ranked_symbol < > > node4(c, {});
	ext::tree < common::ranked_symbol < > > node2(b, {std::move(node3)});
	ext::tree < common::ranked_symbol < > > node1(a, {std::move(node2), std::move(node4)});

	ext::tree < common::ranked_symbol < > > node6(c, {});
	ext::tree < common::ranked_symbol < > > node8(c, {});
	ext::tree < common::ranked_symbol < > > node7(b, {std::move(node8)});
	ext::tree < common::ranked_symbol < > > node5(a, {std::move(node6), std::move(node7)});

	tree::RankedTree < > tree1(alphabet, std::move(node1));
	tree::RankedTree < > tree2(alphabet, std::move(node5));

	CPPUNIT_ASSERT( tree1 != tree2 );
	CPPUNIT_ASSERT( tree1 < tree2 );
	CPPUNIT_ASSERT( tree2 > tree1 );
	CPPUNIT_ASSERT( !(tree1 > tree2) );
}

void TreeTest::testRankedTreeSymbolValidityCheck() {
	const common::ranked_symbol < > a ('a', 2);
	const common::ranked_symbol < > b ('b', 1);
	const common::ranked_symbol < > c ('c', 0);
	const common::ranked_symbol < > d ('d', 0);
	const common::ranked_symbol < > e ('e', 0);

	const ext::set<common::ranked_symbol < >> alphabet {a, b, c, e};

	ext::tree < common::ranked_symbol < > > node3(d, {});
	ext::tree < common::ranked_symbol < > > node4(c, {});
	ext::tree < common::ranked_symbol < > > node2(b, {std::move(node3)});
	ext::tree < common::ranked_symbol < > > node1(a, {std::move(node2), std::move(node4)});

	CPPUNIT_ASSERT_THROW(tree::RankedTree < > (alphabet, std::move(node1)), exception::CommonException);

	ext::tree < common::ranked_symbol < > > node6(e, {});
	ext::tree < common::ranked_symbol < > > node8(c, {});
	ext::tree < common::ranked_symbol < > > node7(b, {std::move(node8)});
	ext::tree < common::ranked_symbol < > > node5(a, {std::move(node6), std::move(node7)});

	tree::RankedTree < > * tree = NULL;
	CPPUNIT_ASSERT_NO_THROW(tree = new tree::RankedTree < > (alphabet, std::move(node5)));

	delete tree;
}

void TreeTest::testUnrankedTreeParser() {
/*
	L=(a, b, c)

	\-1.a
	  |
	  |-2.b
	  | |
	  | \-3.c
	  |
	  \-4.c
*/
	const DefaultSymbolType a = DefaultSymbolType('a');
	const DefaultSymbolType b = DefaultSymbolType('b');
	const DefaultSymbolType c = DefaultSymbolType('c');

	const ext::set<DefaultSymbolType> alphabet {a, b, c};

	ext::tree < DefaultSymbolType > node3(c, {});
	ext::tree < DefaultSymbolType > node4(c, {});
	ext::tree < DefaultSymbolType > node2(b, {std::move(node3)});
	ext::tree < DefaultSymbolType > node1(a, {std::move(node2), std::move(node4)});

	tree::UnrankedTree < > tree(alphabet, std::move(node1));

	CPPUNIT_ASSERT( tree == tree );
	tree.getContent().nicePrint(std::cout);
	{
		ext::deque<sax::Token> tokens = factory::XmlDataFactory::toTokens(tree);
		std::string tmp;
		sax::SaxComposeInterface::printMemory(tmp, tokens);
		std::cout << std::endl << tmp << std::endl << std::endl;

		ext::deque<sax::Token> tokens2;
		sax::SaxParseInterface::parseMemory(tmp, tokens2);
		tree::UnrankedTree < > tree2 = factory::XmlDataFactory::fromTokens (std::move(tokens2));

		CPPUNIT_ASSERT( tree == tree2 );
		std::cout << std::endl;
		tree2.getContent().nicePrint(std::cout);
	}
}

void TreeTest::testUnrankedTreeCompare() {
/*
	L=(a, b, c)

	\-1.a
	  |
	  |-2.b
	  | |
	  | \-3.c
	  |
	  \-4.c

	\-5.a
	  |
	  |-6.c
	  |
	  \-7.b
	    |
	    \-8.c
*/
	const DefaultSymbolType a = DefaultSymbolType('a');
	const DefaultSymbolType b = DefaultSymbolType('b');
	const DefaultSymbolType c = DefaultSymbolType('c');

	const ext::set<DefaultSymbolType> alphabet {a, b, c};

	ext::tree < DefaultSymbolType > node3(c, {});
	ext::tree < DefaultSymbolType > node4(c, {});
	ext::tree < DefaultSymbolType > node2(b, {std::move(node3)});
	ext::tree < DefaultSymbolType > node1(a, {std::move(node2), std::move(node4)});

	ext::tree < DefaultSymbolType > node6(c, {});
	ext::tree < DefaultSymbolType > node8(c, {});
	ext::tree < DefaultSymbolType > node7(b, {std::move(node8)});
	ext::tree < DefaultSymbolType > node5(a, {std::move(node6), std::move(node7)});

	tree::UnrankedTree < > tree1(alphabet, std::move(node1));
	tree::UnrankedTree < > tree2(alphabet, std::move(node5));

	CPPUNIT_ASSERT( tree1 != tree2 );
	CPPUNIT_ASSERT( tree1 < tree2 );
	CPPUNIT_ASSERT( tree2 > tree1 );
	CPPUNIT_ASSERT( !(tree1 > tree2) );
}

void TreeTest::testUnrankedTreeSymbolValidityCheck() {
	const DefaultSymbolType a = DefaultSymbolType('a');
	const DefaultSymbolType b = DefaultSymbolType('b');
	const DefaultSymbolType c = DefaultSymbolType('c');
	const DefaultSymbolType d = DefaultSymbolType('d');
	const DefaultSymbolType e = DefaultSymbolType('e');

	const ext::set<DefaultSymbolType> alphabet {a, b, c, e};

	ext::tree < DefaultSymbolType > node3(d, {});
	ext::tree < DefaultSymbolType > node4(c, {});
	ext::tree < DefaultSymbolType > node2(b, {std::move(node3)});
	ext::tree < DefaultSymbolType > node1(a, {std::move(node2), std::move(node4)});

	CPPUNIT_ASSERT_THROW(tree::UnrankedTree < > (alphabet, std::move(node1)), exception::CommonException);

	ext::tree < DefaultSymbolType > node6(e, {});
	ext::tree < DefaultSymbolType > node8(c, {});
	ext::tree < DefaultSymbolType > node7(b, {std::move(node8)});
	ext::tree < DefaultSymbolType > node5(a, {std::move(node6), std::move(node7)});

	tree::UnrankedTree < > * tree = NULL;
	CPPUNIT_ASSERT_NO_THROW(tree = new tree::UnrankedTree < > (alphabet, std::move(node5)));

	delete tree;
}

void TreeTest::testPrefixRankedBarParser() {
	const DefaultSymbolType bar{alphabet::BarSymbol{}};

	const common::ranked_symbol < > a ('a', 2);
	const common::ranked_symbol < > b ('b', 1);
	const common::ranked_symbol < > c ('c', 0);

	const ext::set<common::ranked_symbol < >> alphabet {a, b, c};

	ext::tree < common::ranked_symbol < > > node3(c, {});
	ext::tree < common::ranked_symbol < > > node4(c, {});
	ext::tree < common::ranked_symbol < > > node2(b, {std::move(node3)});
	ext::tree < common::ranked_symbol < > > node1(a, {std::move(node2), std::move(node4)});

	tree::RankedTree < > tree(alphabet, std::move(node1));
	tree::PrefixRankedBarTree < > tree2(bar, tree);

	CPPUNIT_ASSERT( tree2 == tree2 );
	{
		ext::deque<sax::Token> tokens = factory::XmlDataFactory::toTokens(tree2);
		std::string tmp;
		sax::SaxComposeInterface::printMemory(tmp, tokens);

		std::cout << tmp << std::endl << std::endl;

		ext::deque<sax::Token> tokens2;
		sax::SaxParseInterface::parseMemory(tmp, tokens2);
		tree::PrefixRankedBarTree < > tree3 = factory::XmlDataFactory::fromTokens (std::move(tokens2));

		CPPUNIT_ASSERT( tree2 == tree3 );
		std::cout << std::endl;
	}
}

